#ifndef PROXY_HPP
#define PROXY_HPP

#define WIN32_LEAN_AND_MEAN
#include <WS2tcpip.h>
#include <string>
#include <ostream>
#include <vector>


//// Sockets

typedef unsigned short Port;
typedef unsigned long IP;

bool StartupSockets();
void CleanupSockets();

IP ParseAddress(const char *ip);
std::string AddressToString(IP ip);

bool CreateSocket(int type, int protocol, SOCKET &s);

bool NetBind(SOCKET s, Port port);

Port GetBoundPort(SOCKET s);


std::string SocketGetLastErrorString();
std::string SocketGetErrorString(int code);



class TCPMultiplexer;

class Socket
{
protected:
	TCPMultiplexer *_multiplexer;
	HANDLE _h;
	SOCKET _s;

	bool _ready;
	char *_data_queue;
	int _data_queue_size;

public:
	Socket();
	virtual ~Socket();

	bool IsReady() { return _ready; }
	HANDLE GetHandle() { return _h; }
	SOCKET GetSocket() { return _s; }

	void Initialize(TCPMultiplexer *multiplexer, HANDLE h, SOCKET s);

	bool Post(char *data, int bytes);
	void FlushQueued();

	virtual void OnAccept(SOCKET acceptSocket, sockaddr_in *addr) {}
	virtual void OnConnect() {}
	virtual void OnRead(char *data, int bytes) {}
	virtual void OnClose() {}
};

class FTPParser
{
	char *_buffer;
	int _buffer_size;

public:
	FTPParser();
	virtual ~FTPParser();

protected:
	void AddBuffer(char *buffer, int bytes);

	virtual void OnLine(char *line, int bytes) = 0;
};

class ControlServer : public Socket
{
public:
	IP _target_ip;
	Port _target_port;

public:
	ControlServer(IP target_ip, Port target_port);

	virtual void OnAccept(SOCKET acceptSocket, sockaddr_in *addr);
	virtual void OnClose();
};

class ControlClient;

class ControlConnexion : public Socket, protected FTPParser
{
public:
	ControlClient *_client;
	sockaddr_in _addr;
	IP _target_ip;
	Port _target_port;

	Port CreateDataServer(Port desired_port, Port client_port);

public:
	ControlConnexion(sockaddr_in &addr, IP target_ip, Port target_port);

	virtual void OnRead(char *data, int bytes);
	virtual void OnClose();
	virtual void OnLine(char *line, int bytes);
};

class ControlClient : public Socket, protected FTPParser
{
	ControlConnexion *_connexion;

public:
	ControlClient(ControlConnexion *connexion);

	virtual void OnConnect();
	virtual void OnRead(char *data, int bytes);
	virtual void OnClose();
	virtual void OnLine(char *line, int bytes);
};

class DataServer : public Socket
{
public:
	IP _target_ip;
	Port _target_port;

public:
	DataServer(IP target_ip, Port target_port);

	virtual void OnAccept(SOCKET acceptSocket, sockaddr_in *addr);
	virtual void OnClose();
};

class DataClient;

class DataConnexion : public Socket
{
public:
	DataClient *_client;
	sockaddr_in _addr;
	IP _target_ip;
	Port _target_port;

public:
	DataConnexion(sockaddr_in &addr, IP target_ip, Port target_port);

public:
	virtual void OnRead(char *data, int bytes);
	virtual void OnClose();
};

class DataClient : public Socket
{
public:
	DataConnexion *_connexion;

public:
	DataClient(DataConnexion *connexion);

	virtual void OnConnect();
	virtual void OnRead(char *data, int bytes);
	virtual void OnClose();
};


class TCPMultiplexer
{
protected:
	std::vector<HANDLE> _handles;
	std::vector<Socket*> _sockets;
	std::ostream &_os;

	void WriteLine(unsigned int file_pos, unsigned char *line, int line_bytes, bool c2s);

public:
	bool AddSocket(Socket *s);
	void RemoveSocket(Socket *s);
	void Shutdown();
	void WriteLog(unsigned char *buffer, int bytes, bool c2s);

public:
	TCPMultiplexer(std::ostream &os);
	virtual ~TCPMultiplexer();

	bool Initialize(Port proxy_port, IP target_ip, Port target_port);

	void Spin();
};


#endif // PROXY_HPP
